//
//   File : kvi_http.cpp
//   Creation date : Sat Aug 17 13:43:32 2002 GMT by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "kvi_http.h"
#include "kvi_locale.h"
#include "kvi_netutils.h"
#include "kvi_dns.h"
#include "kvi_error.h"
#include "kvi_debug.h"
#include "kvi_socket.h"

#include <qtimer.h>

KviHttpRequest::KviHttpRequest()
: QObject()
{
	m_pDns = 0;
	m_pThread = 0;
	m_szLastError = __tr("No request");
	m_pFile = 0;
}

KviHttpRequest::~KviHttpRequest()
{
	reset();
}

void KviHttpRequest::reset()
{
	m_bHeaderProcessed = false;
	m_szBuffer = "";
	m_szLastError = __tr("No request");
	m_szHeader = "";
	m_szResponse = "";

	if(m_pDns)delete m_pDns;
	if(m_pThread)delete m_pThread;
	KviThreadManager::killPendingEvents(this);
	if(m_pFile)
	{
		m_pFile->close();
		delete m_pFile;
	}
	m_pFile = 0;
	m_pDns = 0;
	m_pThread = 0;
}

bool KviHttpRequest::get(const KviUrl &u,ProcessingType p,const char * szFileName)
{
	reset();
	m_url = u;
	m_processingType = p;

	if(m_processingType == StoreToFile)
	{
		KviStr fileName = szFileName;
		if(fileName.isEmpty())
		{
			m_szLastError = __tr("No filename specified for the \"StoreToFile\" processing type");
			return false;
		}

		m_pFile = new QFile(fileName.ptr());
		if(!m_pFile->open(IO_WriteOnly))
		{
			delete m_pFile;
			m_szLastError.sprintf(__tr("Can't open the file \"%s\" for writing"),szFileName);
			return false;
		}
	}

	if(m_url.host().isEmpty())
	{
		m_szLastError = __tr("Invalid url: missing host name");
		return false;
	}

	if(!kvi_strEqualCI(m_url.protocol().ptr(),"http"))
	{
		m_szLastError.sprintf(__tr("Unsupported protocol %s"),m_url.protocol().ptr());
		return false;
	}

	if(kvi_isValidStringIp(m_url.host().ptr()))
	{
		m_szIp = m_url.host();
		QTimer::singleShot(10,this,SLOT(haveServerIp()));
		return true;
	}

	return startDnsLookup();
}

bool KviHttpRequest::startDnsLookup()
{
	m_pDns = new KviDns();
	connect(m_pDns,SIGNAL(lookupDone(KviDns *)),this,SLOT(dnsLookupDone(KviDns *)));

	if(!m_pDns->lookup(m_url.host().ptr(),KviDns::IpV4))
	{
		delete m_pDns;
		m_pDns = 0;
		m_szLastError = __tr("Can't start the dns lookup");
		return false;
	}

	KviStr tmp(KviStr::Format,__tr("Looking up host %s"),m_url.host().ptr());
	emit status(tmp.ptr());

	return true;
}

void KviHttpRequest::dnsLookupDone(KviDns *d)
{
	__range_valid(m_pDns == d);

	if(d->state() == KviDns::Success)
	{
		m_szIp = d->firstIpAddress();
		delete m_pDns;
		m_pDns = 0;
		KviStr tmp(KviStr::Format,__tr("Host %s resolved to %s"),m_url.host().ptr(),m_szIp.ptr());
		emit status(tmp.ptr());
		haveServerIp();
	} else {
		m_szLastError = kvi_getErrorString(d->error());
		delete m_pDns;
		m_pDns = 0;
		emit terminated(false);
	}
}

void KviHttpRequest::haveServerIp()
{
	if(m_pThread)delete m_pThread;
	m_pThread = new KviHttpRequestThread(this,m_szIp.ptr(),m_url.path().ptr());
	if(!m_pThread->start())
	{
		delete m_pThread;
		m_pThread = 0;
		m_szLastError = __tr("Can't start the request slave thread");
		emit terminated(false);
		return;
	}
	KviStr tmp(KviStr::Format,__tr("Contacting host %s on port %u"),m_szIp.ptr(),m_url.port());
	emit status(tmp.ptr());
}

bool KviHttpRequest::event(QEvent *e)
{
	if(e->type() == KVI_THREAD_EVENT)
	{
		switch(((KviThreadEvent *)e)->id())
		{
			case KVI_THREAD_EVENT_SUCCESS:
				switch(m_processingType)
				{
					case WholeFile:
						emit data(m_szBuffer);
					break;
					case Lines:
						if(m_szBuffer.hasData())emit data(m_szBuffer);
					break;
					case StoreToFile:
						// never happens:if(m_szBuffer.hasData())m_pFile->writeBlock(m_szBuffer.ptr(),m_szBuffer.len());
						m_pFile->close();
					break;
				}
				reset();
				emit terminated(true);
				return true;
			break;
			case KVI_THREAD_EVENT_ERROR:
			{
				KviStr * err = ((KviThreadDataEvent<KviStr> *)e)->getData();
				m_szLastError = __tr_no_xgettext(err->ptr());
				emit terminated(false);
				delete err;
				reset();
				return true;
			}
			break;
			case KVI_THREAD_EVENT_DATA:
			{
				KviStr * d = ((KviThreadDataEvent<KviStr> *)e)->getData();
				processData(d);
				delete d;
				return true;
			}
			break;
			case KVI_THREAD_EVENT_MESSAGE:
			{
				KviStr * msg = ((KviThreadDataEvent<KviStr> *)e)->getData();
				emit status(__tr_no_xgettext(msg->ptr()));
				delete msg;
				return true;
			}
			break;
		}
	}
	return QObject::event(e);
}

void KviHttpRequest::emitLines()
{
	int idx=m_szBuffer.findFirstIdx('\n');
	while(idx != -1)
	{
		KviStr tmp = m_szBuffer.left(idx);
		tmp.stripRight('\r');
		emit data(tmp);
		m_szBuffer.cutLeft(idx + 1);
		idx = m_szBuffer.findFirstIdx('\n');
	}
}

void KviHttpRequest::processData(KviStr * d)
{
	if(m_bHeaderProcessed)
	{
		// we're just getting data 
		switch(m_processingType)
		{
			case Blocks:
				emit data(*d);
			break;
			case Lines:
				m_szBuffer.append(*d);
				emitLines();
			break;
			case StoreToFile:
				m_pFile->writeBlock(d->ptr(),d->len());
			break;
			default:
				m_szBuffer.append(*d);
			break;
		}
	} else {
		m_szBuffer.append(*d);
		int idx = m_szBuffer.findFirstIdx("\r\n\r\n");
		if(idx != -1)
		{
			m_szHeader = m_szBuffer.left(idx);
			m_szBuffer.cutLeft(idx + 4);
			idx = m_szHeader.findFirstIdx("\r\n");
			if(idx != -1)
			{
				m_szResponse = m_szHeader.left(idx);
			} else {
				m_szResponse = m_szHeader;
			}

			KviStr tmp(KviStr::Format,__tr("Received HTTP response: %s"),m_szResponse.ptr());
			emit status(tmp.ptr());
			m_bHeaderProcessed = true;
			switch(m_processingType)
			{
				case Blocks:
					if(m_szBuffer.hasData())emit data(m_szBuffer);
					m_szBuffer = "";
				break;
				case Lines:
					if(m_szBuffer.hasData())emitLines();
				break;
				case StoreToFile:
					m_pFile->writeBlock(m_szBuffer.ptr(),m_szBuffer.len());
					m_szBuffer = "";
				break;
			}
		}
	}
}
















KviHttpRequestThread::KviHttpRequestThread(KviHttpRequest * r,const char * szIp,const char * szPath)
: KviSensitiveThread()
{
	m_pRequest = r;
	m_szIp = szIp;
	m_szPath = szPath;
}

KviHttpRequestThread::~KviHttpRequestThread()
{
}

bool KviHttpRequestThread::processInternalEvents()
{
	while(KviThreadEvent *e = dequeueEvent())
	{
		switch(e->id())
		{
			case KVI_THREAD_EVENT_TERMINATE:
			{
				delete e;
				return false;
			}
			break;
			default:
				debug("Unrecognized event in http thread");
				delete e;
				return false;
			break;
		}
	}
	return true;
}

bool KviHttpRequestThread::failure(const char * error)
{
	if(m_sock != KVI_INVALID_SOCKET)kvi_socket_close(m_sock);
	m_sock = KVI_INVALID_SOCKET;

	if(error)
	{
		postEvent(m_pRequest,new KviThreadDataEvent<KviStr>(KVI_THREAD_EVENT_ERROR,new KviStr(error)));
	} else {
		postEvent(m_pRequest,new KviThreadDataEvent<KviStr>(KVI_THREAD_EVENT_ERROR,new KviStr(__tr_no_lookup("Aborted"))));
	}

	return false;
}


bool KviHttpRequestThread::selectForWrite(int iTimeoutInSecs)
{

	time_t startTime = time(0);

	for(;;)
	{
		if(!processInternalEvents())
		{
			return failure();
		}

		fd_set writeSet;
	
		FD_ZERO(&writeSet);

		FD_SET(m_sock,&writeSet);

		struct timeval tmv;
		tmv.tv_sec  = 0;
		tmv.tv_usec = 1000; // we wait 1000 usecs for an event
	

		int nRet = kvi_socket_select(m_sock + 1,0,&writeSet,0,&tmv);
	
		if(nRet > 0)
		{
			if(FD_ISSET(m_sock,&writeSet))
			{
				// connected!
				return true;
			}
		} else {
			if(nRet < 0)
			{
				int err = kvi_socket_error();
#ifdef COMPILE_ON_WINDOWS
				if((err != EAGAIN) && (err != EINTR) && (err != WSAEWOULDBLOCK))
#else
				if((err != EAGAIN) && (err != EINTR))
#endif
				{
					KviStr tmp(KviStr::Format,__tr_no_lookup("Select error: %s (errno=%d)"),
						kvi_getErrorString(kvi_errorFromSystemError(err)),err);
					return failure(tmp.ptr());
				}
			}
		}


		if((time(0) - startTime) > iTimeoutInSecs)return failure(__tr_no_lookup("Operation timed out"));

		usleep(100000); // 1/10 sec
	}

	return false;
}

bool KviHttpRequestThread::connectToRemoteHost()
{
	m_sock = kvi_socket_create(KVI_SOCKET_PF_INET,KVI_SOCKET_TYPE_STREAM,0); //tcp
	if(m_sock == KVI_INVALID_SOCKET)
		return failure(__tr_no_lookup("Failed to create the socket"));

	if(!kvi_socket_setNonBlocking(m_sock))
		return failure(__tr_no_lookup("Failed to enter non blocking mode"));

	sockaddr_in saddr;

	if(!kvi_stringIpToBinaryIp(m_szIp.ptr(),&(saddr.sin_addr)))
		return failure(__tr_no_lookup("Invalid target address"));

	saddr.sin_port = htons(80);
	saddr.sin_family = AF_INET;

	if(!kvi_socket_connect(m_sock,(struct sockaddr *)&saddr,sizeof(saddr)))
	{
		int err = kvi_socket_error();
		if(!kvi_socket_recoverableConnectError(err))
		{
			KviStr tmp(KviStr::Format,__tr_no_lookup("Connect error: %s (errno=%d)"),
				kvi_getErrorString(kvi_errorFromSystemError(err)),err);
			return failure(tmp.ptr());
		}
	}

	// now loop selecting for write

	//#warning "This should be a tuneable timeout"

	if(!selectForWrite(60))return false;

	int sockError;
	int iSize=sizeof(sockError);
	if(!kvi_socket_getsockopt(m_sock,SOL_SOCKET,SO_ERROR,(void *)&sockError,&iSize))sockError = -1;
	if(sockError != 0)
	{
		//failed
		if(sockError > 0)sockError = kvi_errorFromSystemError(sockError);
		else sockError = KviError_unknownError;
		KviStr tmp(KviStr::Format,__tr_no_lookup("Connect error: %s (errno=%d)"),
			kvi_getErrorString(sockError),sockError);
		return failure(tmp.ptr());
	}

	return true;
}


bool KviHttpRequestThread::sendBuffer(const char * buffer,int bufLen,int iTimeoutInSecs)
{
	const char * ptr = buffer;
	int curLen       = bufLen;

	time_t startTime = time(0);

	for(;;)
	{
		if(!processInternalEvents())return failure();

		int wrtn = kvi_socket_send(m_sock,ptr,curLen);
		if(wrtn > 0)
		{
			curLen -= wrtn;

			if(curLen <= 0)break;

			ptr += wrtn;
		} else {
			if(wrtn < 0)
			{
				int err = kvi_socket_error();
#ifdef COMPILE_ON_WINDOWS
				if((err != EAGAIN) && (err != EINTR) && (err != WSAEWOULDBLOCK))
#else
				if((err != EAGAIN) && (err != EINTR))
#endif
				{
					KviStr tmp(KviStr::Format,__tr_no_lookup("Write error: %s (errno=%d)"),
							kvi_getErrorString(kvi_errorFromSystemError(err)),err);
					return failure(tmp.ptr());
				}
			}
		}

		int diff = time(0) - startTime;
		if(diff > iTimeoutInSecs)
			return failure(__tr_no_lookup("Operation timed out"));

		usleep(10000);
	}

	return true;
}


int KviHttpRequestThread::selectForReadStep()
{
	// calls select on the main socket
	// returns 1 if there is data available for reading
	// returns 0 if there is no data available but there was no error
	// returns -1 if there was a critical error (socket closed)
	fd_set readSet;

	FD_ZERO(&readSet);

	FD_SET(m_sock,&readSet);

	struct timeval tmv;
	tmv.tv_sec  = 0;
	tmv.tv_usec = 1000; // we wait 1000 usecs for an event


	int nRet = kvi_socket_select(m_sock + 1,&readSet,0,0,&tmv);

	if(nRet > 0)
	{
		if(FD_ISSET(m_sock,&readSet))
		{
			// ok
			return 1;
		}
	} else {
		if(nRet < 0)
		{
			int err = kvi_socket_error();
#ifdef COMPILE_ON_WINDOWS
			if((err != EAGAIN) && (err != EINTR) && (err != WSAEWOULDBLOCK))
#else
			if((err != EAGAIN) && (err != EINTR))
#endif
			{
				KviStr tmp(KviStr::Format,__tr_no_lookup("Select error: %s (errno=%d)"),
					kvi_getErrorString(kvi_errorFromSystemError(err)),err);
				failure(tmp.ptr());
				return -1;
			}
		}
	}

	return 0;
}


bool KviHttpRequestThread::selectForRead(int iTimeoutInSecs)
{
	// waits for some data to arrive on the socket
	// up to iTimeoutInSecs seconds
	// returns true if data is available on the socket
	// or false if there was a select() error or no data
	// was available in the specified amount of time

	time_t startTime = time(0);

	for(;;)
	{
		if(!processInternalEvents())
		{
			return failure(); // ensure that the socket is closed
		}

		int nRet = selectForReadStep();

		if(nRet < 0)return false;
		if(nRet > 0)return true;

		int diff = time(0) - startTime;
		if(diff > iTimeoutInSecs)
			return failure(__tr_no_lookup("Operation timed out (while selecting for read)"));

		usleep(100000); // 1/10 sec
	}

	return false;
}

bool KviHttpRequestThread::readData()
{
	char buffer[1025];

	int readed = kvi_socket_read(m_sock,buffer,1024);
	if(readed > 0)
	{
		buffer[readed] = '\0';
		postEvent(m_pRequest,new KviThreadDataEvent<KviStr>(KVI_THREAD_EVENT_DATA,new KviStr(buffer)));
	} else {
		if(readed < 0)
		{
			// Read error ?
			int err = kvi_socket_error();
#ifdef COMPILE_ON_WINDOWS
			if((err != EAGAIN) && (err != EINTR) && (err != WSAEWOULDBLOCK))
#else
			if((err != EAGAIN) && (err != EINTR))
#endif
			{
				// yes...read error
				KviStr tmp(KviStr::Format,__tr_no_lookup("Read error: %s (errno=%d)"),
						kvi_getErrorString(kvi_errorFromSystemError(err)),err);
				return failure(tmp.ptr());
			}
			return true; // EINTR or EAGAIN...transient problem
		} else {
			// readed == 0
			// Connection closed by remote host
			kvi_socket_close(m_sock);
			postEvent(m_pRequest,new KviThreadEvent(KVI_THREAD_EVENT_SUCCESS));
			return false;
		}
	}
	return true;
}


void KviHttpRequestThread::run()
{
	if(!connectToRemoteHost())return;

	postEvent(m_pRequest,new KviThreadDataEvent<KviStr>(KVI_THREAD_EVENT_MESSAGE,new KviStr(__tr_no_lookup("Connection estabilished: sending HTTP request"))));

	KviStr szRequest(KviStr::Format,
		"GET %s HTTP/1.0\r\n" \
		"Connection: Keep-Alive\r\n" \
		"User-Agent: KVIrc-http-slave-1.0.0\r\n\r\n",m_szPath.ptr());

	if(!sendBuffer(szRequest.ptr(),szRequest.len(),60))return;
	// now loop reading data

	postEvent(m_pRequest,new KviThreadDataEvent<KviStr>(KVI_THREAD_EVENT_MESSAGE,new KviStr(__tr_no_lookup("Receiving data ..."))));

	for(;;)
	{
		if(!selectForRead(180))return;
		if(!readData())return;
	}
}

#include "kvi_http.moc"
