//
//   File : libkviregchan.cpp
//   Creation date : Sat Jun 29 02:55:18 2002 GMT by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 2001 Szymon Stefanek (stefanek@tin.it)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "kvi_module.h"
#include "kvi_uparser.h"
#include "kvi_command.h"
#include "kvi_regchan.h"
#include "kvi_locale.h"
#include "kvi_out.h"
#include "kvi_mirccntrl.h"
#include "kvi_window.h"

// kvi_app.cpp
extern KVIRC_API KviRegisteredChannelDataBase * g_pRegisteredChannelDataBase;

/*
	@doc: regchan.add
	@type:
		command
	@title:
		regchan.add
	@keyterms:
		registering channels
	@short:
		Registers a channel
	@syntax:
		regchan.add <channel> [netmask]
	@description:
		Registers the <channel> with the specified [netmask].<br>
		If netmask is not passed a netmask of * is assumed.<br>
	@examples:
		[example]
			reguser.add #kvirc openprojects
		[/example]
	@seealso:
*/

static bool regchan_module_cmd_add(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"regchan_module_cmd_add");

	KviStr name,mask;
	if(!g_pUserParser->parseCmdSingleToken(c,name))return false;
	if(!g_pUserParser->parseCmdFinalPart(c,mask))return false;

	if(name.isEmpty())
	{
		c->warning(__tr("No name specified"));
		return c->leaveContext();
	}

	if(mask.isEmpty())mask="*";

	g_pRegisteredChannelDataBase->add(new KviRegisteredChannel(name,mask));
	return c->leaveContext();
}

/*
	@doc: regchan.remove
	@type:
		command
	@title:
		regchan.remove
	@keyterms:
		unregistering channels
	@short:
		Unregisters a channel
	@syntax:
		regchan.remove [-q] <channel> <netmask>
	@description:
		Unregisters the <channel> that has <netmask>.[br]
		If no exact <channel>/<netmask> is found in the database then
		a warning is printed unless -q is used.[br]
	@examples:
		[example]
			reguser.remove #kvirc openprojects
		[/example]
	@seealso:
		regchan.add
*/

static bool regchan_module_cmd_remove(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"regchan_module_cmd_remove");

	KviStr name,mask;
	if(!g_pUserParser->parseCmdSingleToken(c,name))return false;
	if(!g_pUserParser->parseCmdFinalPart(c,mask))return false;

	if(name.isEmpty())
	{
		c->warning(__tr("No name specified"));
		return c->leaveContext();
	}

	if(mask.isEmpty())
	{
		c->warning(__tr("No netmask specified"));
		return c->leaveContext();
	}

	KviRegisteredChannel * ch = g_pRegisteredChannelDataBase->findExact(name,mask);
	if(ch)
	{
		g_pRegisteredChannelDataBase->remove(ch);
	} else {
		if(!c->hasSwitch('q'))c->warning(__tr("No such channel/netmask entry in the database"));
	}
	return c->leaveContext();
}



/*
	@doc: regchan.setproperty
	@type:
		command
	@title:
		regchan.setproperty
	@short:
		Sets a property for a registered channel
	@syntax:
		regchan.setproperty [-q] <channel> <netmask> <property> <value>
	@description:
		Sets the <property> for the <channel> with <netmask> to <value>.[br]
		If <value> is an empty string the <property> is removed.[br]
		If no exact <channel>/<netmask> is found in the database then
		a warning is printed unless -q is used.[br]
	@examples:
		[example]
			reguser.setproperty #kvirc openprojects autojoin true
		[/example]
	@seealso:
		regchan.add
*/

static bool regchan_module_cmd_setproperty(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"regchan_module_cmd_setproperty");

	KviStr name,mask,prop,value;
	if(!g_pUserParser->parseCmdSingleToken(c,name))return false;
	if(!g_pUserParser->parseCmdSingleToken(c,mask))return false;
	if(!g_pUserParser->parseCmdSingleToken(c,prop))return false;
	if(!g_pUserParser->parseCmdFinalPart(c,value))return false;

	if(name.isEmpty())
	{
		c->warning(__tr("No name specified"));
		return c->leaveContext();
	}

	if(mask.isEmpty())
	{
		c->warning(__tr("No netmask specified"));
		return c->leaveContext();
	}

	if(prop.isEmpty())
	{
		c->warning(__tr("No property name specified"));
		return c->leaveContext();
	}


	KviRegisteredChannel * ch = g_pRegisteredChannelDataBase->findExact(name,mask);
	if(ch)
	{
		if(value.hasData())
		{
			ch->setProperty(prop.ptr(),new KviStr(value));
		} else {
			ch->removeProperty(prop.ptr());
		}
	} else {
		if(!c->hasSwitch('q'))c->warning(__tr("No such channel/netmask entry in the database"));
	}
	return c->leaveContext();
}



/*
	@doc: regchan.list
	@type:
		command
	@title:
		regchan.list
	@short:
		Lists the registered channels
	@syntax:
		regchan.list
	@description:
		Lists the registered channels with their description
*/

static bool regchan_module_cmd_list(KviModule *m,KviCommand *c)
{
	ENTER_CONTEXT(c,"regchan_module_cmd_list");

	KviStr dummy;
	if(!g_pUserParser->parseCmdFinalPart(c,dummy))return false;

	c->window()->output(KVI_OUT_SYSTEMMESSAGE,__tr("Registered channel list:"));

	int tot = 0;

	QAsciiDictIterator<KviRegisteredChannelList> it(*(g_pRegisteredChannelDataBase->channelDict()));
	while(KviRegisteredChannelList * l = it.current())
	{
		for(KviRegisteredChannel * ch = l->first();ch;ch = l->next())
		{
			tot++;
			c->window()->output(KVI_OUT_SYSTEMMESSAGE,__tr("Channel: %c%s@%s"),
				KVI_TEXT_BOLD,ch->name().ptr(),ch->netMask().ptr());
			QAsciiDictIterator<KviStr> pit(*(ch->propertyDict()));
			while(KviStr * s = pit.current())
			{
				c->window()->output(KVI_OUT_SYSTEMMESSAGE,__tr("   Property: %s=%s"),pit.currentKey(),s->ptr());
				++pit;
			}
		}
		++it;
	}

	c->window()->output(KVI_OUT_SYSTEMMESSAGE,__tr("Total %d channels"),tot);
	return c->leaveContext();
}


/*
	@doc: regchan.property
	@type:
		function
	@title:
		$regchan.property
	@short:
		Returns a property of an entry in the registered channels database
	@syntax:
		$regchan.property(<channel_name>,<network_mask>,<property_name>)
	@description:
		Returns the value of the property identified by <property_name> and
		bound to the channel identified by <channel_name> and matching <network_mask>.[br]
		If the property is not set or if there is no matching channel entry an empty string is returned.[br]
		<network_mask> may contain * and ? wildcards.
	@examples:
		[example]
			echo $regchan.property(#kvirc,*,autojoin)
		[/example]
*/

static bool regchan_module_fnc_property(KviModule *m,KviCommand *c,KviParameterList * parms,KviStr &buffer)
{
	ENTER_CONTEXT(c,"regchan_module_fnc_property");
	KviRegisteredChannel * ch = g_pRegisteredChannelDataBase->find(parms->safeFirstParam(),parms->safeNextParam());
	if(ch)
	{
		const char * n = parms->safeNextParam();
		if(n)
		{
			KviStr * p = ch->property(n);
			if(p)buffer.append(p->ptr());
		} else c->warning(__tr("Missing property name"));
	} //else c->warning(__tr("User %s not found"),parms->safeFirstParam());
	return c->leaveContext();
}


static bool regchan_module_init(KviModule * m)
{
	m->registerCommand("add",regchan_module_cmd_add);
	m->registerCommand("list",regchan_module_cmd_list);
	m->registerCommand("remove",regchan_module_cmd_remove);
	m->registerCommand("setproperty",regchan_module_cmd_setproperty);
	m->registerFunction("property",regchan_module_fnc_property);
	return true;
}

static bool regchan_module_cleanup(KviModule *m)
{
	return true;
}

static bool regchan_module_can_unload(KviModule *)
{
	return true;
}

KVIMODULEEXPORTDATA KviModuleInfo kvirc_module_info=
{
	"RegChan",                                              // module name
	"1.0.0",                                                // module version
	"Copyright (C) 2002 Szymon Stefanek (pragma at kvirc dot net)", // author & (C)
	"Script interface to the registered channels database",
	regchan_module_init,
	regchan_module_can_unload,
	0,
	regchan_module_cleanup
};
